#![allow(unused)]
use core::fmt::{self, Debug, Formatter};
use ring::{aead::Nonce, rand};
use serde::{Deserialize, Serialize};
use std::{
    fs,
    fs::File as StdFile,
    io,
    path::{Path, PathBuf},
};

pub mod util;
use util::{
    encrypt_bytes, path_push, random, random_usize, read_json_from_file, write_json_to_file,
    CryptoError,
};

type CryptoResult<O> = Result<O, CryptoError>;

/// Location of the super block, maybe change to LUKS2fs in the future
const METADATA_PATH: &str = "encfsrs.metadata";

/// Initialises a directory, will not overwrite existing setup.
/// Gets called automatically if `open_dir` encounters an empty directory.
pub fn init_dir<P: AsRef<Path>, S: AsRef<str>>(
    base_path: P,
    paper_key: S,
) -> CryptoResult<CryptoContext> {
    let base_path: &Path = base_path.as_ref();
    std::fs::create_dir_all(base_path)?;
    if base_path.is_dir() == false {
        return Err(CryptoError::InvalidBackend);
    }
    let superblock_path = path_push(base_path, METADATA_PATH);
    if superblock_path.exists() {
        return Err(CryptoError::BackendAlreadyExists);
    }

    let ctx = CryptoContext::random_init(base_path.to_owned(), paper_key)?;

    ctx.store_to_disk()?;

    Ok(ctx)
}

pub fn open_dir<P: AsRef<Path>, S: AsRef<str>>(
    base_path: P,
    key: S,
) -> CryptoResult<CryptoContext> {
    let base_path: &Path = base_path.as_ref();
    if base_path.is_dir() == false {
        return Err(CryptoError::InvalidBackend);
    }

    let superblock = path_push(base_path, METADATA_PATH);
    if superblock.exists() == false {
        return init_dir(base_path, key);
    }
    let superblock: SuperBlock = read_json_from_file(superblock)?;

    todo!()
}

/// Holds encrypted keys for the volume on disk
#[derive(Serialize, Deserialize, PartialEq, Debug)]
pub struct SuperBlock {
    /// These keys are stored either on
    /// * Paper
    /// * TPM
    /// * On-Disk or Code (worst case)
    encrypted_opening_keys: Vec<Vec<u8>>,
    /// This key is used for all the actual encryption.
    /// It must not be exposed to the user
    //encrypted_root_key: Vec<u8>,
    encrypted_nonce: Vec<u8>,
}

/// Is the decrypted root key
pub struct MasterKey([u8; 32]);

/// Holds the decrypted root key
//#[derive(Debug)]
pub struct CryptoContext {
    /// This counter must never be reused
    nonce_counter: Nonce,
    secret_key: MasterKey,
    opening_keys: Vec<ContextKey>,
    /// Where in the unencrypted filesystem is the store
    base_path: PathBuf,
}
impl CryptoContext {
    pub fn random_init<S: AsRef<str>>(base_path: PathBuf, paper_key: S) -> CryptoResult<Self> {
        use ring::rand::SecureRandom;
        let mut rng = ring::rand::SystemRandom::new();

        let mut secret_key = [0u8; 32];
        rng.fill(&mut secret_key)?;

        let mut nonce_counter = [0u8; 12];
        nonce_counter[11] = 1;

        Ok(CryptoContext {
            nonce_counter: Nonce::try_assume_unique_for_key(&nonce_counter)?,
            secret_key: MasterKey(secret_key),
            opening_keys: vec![ContextKey::from_paper_key(paper_key.as_ref())],
            base_path,
        })
    }
    /// Encrypt the current Context for later use
    fn store_to_disk(&self) -> CryptoResult<()> {
        let master_key = &self.secret_key;
        let superblock = SuperBlock {
            encrypted_opening_keys: self
                .opening_keys
                .iter()
                .map(|k| k.encrypt(master_key))
                .collect(),
            //encrypted_root_key:
            encrypted_nonce: encrypt_bytes(self.nonce_counter.as_ref(), master_key),
        };
        let mut superblock_path = self.base_path.clone();
        superblock_path.push(METADATA_PATH);
        write_json_to_file(&superblock, superblock_path)?;
        Ok(())
    }
}
impl Debug for CryptoContext {
    fn fmt(&self, fmt: &mut Formatter) -> fmt::Result {
        fmt.write_str("CryptoContext { ... }")
    }
}

pub struct ContextKey {
    key_material: Vec<u8>,
}
impl ContextKey {
    fn from_paper_key(paper: &str) -> Self {
        todo!("key strengthening with Argon2")
    }
    fn from_binary_key(binary: &[u8]) -> Self {
        todo!("decryption from TPM or similar")
    }
    fn encrypt(&self, master_key: &MasterKey) -> Vec<u8> {
        todo!("encrypt and hash")
    }
}

pub struct File;

#[cfg(test)]
mod tests {
    use super::*;
    use std::fs::{create_dir_all, read_to_string};

    #[test]
    fn invalid_backend() {
        assert_eq!(
            CryptoError::InvalidBackend,
            open_dir("/dev/null", "KEY").unwrap_err()
        );
    }
    #[test]
    fn double_backend_init() {
        let temp = &temp_dir();
        create_dir_all(&temp).unwrap();
        write(temp, METADATA_PATH, "does not matter").unwrap();

        assert_eq!(
            CryptoError::BackendAlreadyExists,
            init_dir(temp, "KEY").unwrap_err()
        );
    }
    #[test]
    fn open_empty_dir() {
        let temp = &temp_dir();
        create_dir_all(&temp).unwrap();

        let ctx = open_dir(&temp, "KEY").unwrap();
        drop(ctx);

        assert!(path_push(&temp, METADATA_PATH).exists())
    }

    #[test]
    fn read_superblock() {
        let temp = &temp_dir();

        create_dir_all(&temp).unwrap();
        let superblock = write(
            temp,
            METADATA_PATH,
            r#"{
                "encrypted_opening_keys": [],
                "encrypted_nonce": [1,2,3,4,5,6,7,8,9,10,11,12]
            }"#,
        )
        .unwrap();

        let e = SuperBlock {
            encrypted_opening_keys: vec![],
            encrypted_nonce: vec![1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
        };

        assert_eq!(e, read_json_from_file(superblock).unwrap());
    }

    fn temp_dir() -> temp_testdir::TempDir {
        let destroy = false;
        let base_path = env!("OUT_DIR");
        //std::env::var("OUT_DIR").expect("cargo OUT_DIR target dir not set")
        let mut base_path = PathBuf::from(base_path);
        let r: usize = random_usize();
        base_path.push(format!("encfsstore-test-{r}"));
        temp_testdir::TempDir::new(base_path, destroy)
    }

    fn write<P: AsRef<Path>>(
        temp: &temp_testdir::TempDir,
        path: P,
        data: &str,
    ) -> std::io::Result<PathBuf> {
        let mut file_path = PathBuf::from(temp.as_ref());
        file_path.push(path);
        println!("test::write({})", file_path.display());

        std::fs::write(&file_path, data)?;
        Ok(file_path)
    }
}
